import datetime
import io
import logging
import re
import unittest
from itertools import chain
from typing import Any, Dict, List
from unittest.mock import ANY, patch

import yaml

from prometheus_icinga_exporter.alertmanager_client import PostableAlert
from prometheus_icinga_exporter.am import (
    PrometheusIcingaAM,
    TeamRegexes,
    TeamTagMatcher,
    UnknownLabel,
)
from prometheus_icinga_exporter.am import __name__ as PrometheusIcingaExporterModuleName
from prometheus_icinga_exporter.am import main

service_minimal = {
    "host_name": "myhost",
    "service_description": "a problem",
    "status_information": "CRITICAL - Socket timeout after 10 seconds",
    "status": "CRITICAL",
}

service_critical = {
    "host_name": "dummy",
    "host_display_name": "dummy",
    "service_description": "a problem",
    "service_display_name": "a problem",
    "status": "CRITICAL",
    "last_check": "2020-09-17 08:30:15",
    "duration": "52d  0h 17m 18s",
    "attempts": "3/3",
    "current_notification_number": 1,
    "state_type": "HARD",
    "is_flapping": False,
    "in_scheduled_downtime": False,
    "active_checks_enabled": True,
    "passive_checks_enabled": True,
    "notifications_enabled": True,
    "has_been_acknowledged": False,
    "action_url": None,
    "notes_url": "https://url.here",
    "status_information": "CRITICAL - Socket timeout after 10 seconds",
}


class StatusParsing(unittest.TestCase):
    def setUp(self):
        notes_map = {
            re.compile("grafana"): "dashboard",
            re.compile("wiki"): "runbook",
            re.compile("logstash"): "logs",
        }
        self.am = PrometheusIcingaAM(
            ["http://localhost:42"],
            "query_string",
            "/path/status.cgi",
            notes_map=notes_map,
        )

    def _status_from_service(self, service):
        """Take a service_status list entry and return a fully-formed
        'status' object"""
        return {"status": {"service_status": [service]}}

    def test_invalid(self):
        self.assertRaises(KeyError, self.am.alerts_from_status_json, {})

    def test_critical(self):
        status = self._status_from_service(service_critical)
        problems = self.am.alerts_from_status_json(status)
        self.assertEqual(1, len(problems))

    def test_annotations_url(self):
        t = {"notes_url": "http://foo.bar"}
        t.update(service_minimal)

        status = self._status_from_service(t)
        problems = self.am.alerts_from_status_json(status)
        self.assertEqual(1, len(problems))
        self.assertEqual(problems[0].annotations["link"], "http://foo.bar")

    def test_annotations_urls(self):
        t = {"notes_url": "'http://bar.baz' 'http://foo.bar'"}
        t.update(service_minimal)

        status = self._status_from_service(t)
        p = self.am.alerts_from_status_json(status)[0]
        self.assertEqual(p.annotations["link"], "http://bar.baz")
        self.assertEqual(p.annotations["link1"], "http://foo.bar")

    def test_annotations_map(self):
        t = {"notes_url": "'http://grafana' 'http://wiki' 'http://grafana'"}
        t.update(service_minimal)

        status = self._status_from_service(t)
        p = self.am.alerts_from_status_json(status)[0]
        self.assertEqual(p.annotations["dashboard"], "http://grafana")
        self.assertEqual(p.annotations["dashboard1"], "http://grafana")
        self.assertEqual(p.annotations["runbook"], "http://wiki")

    def test_parse_cgi_json(self):
        self.assertIsNone(self.am._parse_cgi_json(b""))
        self.assertEqual({}, self.am._parse_cgi_json(b"foo: bar\r\n\r\n{}"))

    def test_quote_url(self):
        cases = {
            "http://foo.bar.baz/path/app?param=foo bar&bar=baz": "http://foo.bar.baz/path/app?param=foo+bar&bar=baz",
            "http://foo.bar.baz/path/app": "http://foo.bar.baz/path/app",
        }

        for raw, expected in cases.items():
            self.assertEqual(self.am._quote_url(raw), expected)


class GeneratorURL(unittest.TestCase):
    def setUp(self):
        self.am = PrometheusIcingaAM(
            ["http://localhost:42"],
            "query_string",
            "/path/status.cgi",
            extinfo_url="https://dum.my/cgi-bin/extinfo.cgi",
        )

    def _status_from_service(self, service):
        """Take a service_status list entry and return a fully-formed
        'status' object"""
        return {"status": {"service_status": [service]}}

    def test_generator_url(self):
        t = {"notes_url": "http://foo.bar"}
        t.update(service_minimal)

        status = self._status_from_service(t)
        problems = self.am.alerts_from_status_json(status)
        self.assertEqual(1, len(problems))
        self.assertEqual(
            problems[0].generator_url,
            "https://dum.my/cgi-bin/extinfo.cgi?type=2&host=myhost&service=a+problem",
        )


class AlertsList(unittest.TestCase):
    @patch("prometheus_icinga_exporter.am.PrometheusIcingaAM._get_current_alerts")
    def test_resolved(self, current_alerts_mock):
        am = PrometheusIcingaAM(
            ["http://localhost:42"], "query_string", "/path/status.cgi"
        )
        current_alerts_mock.return_value = [
            PostableAlert(labels={"foo": "bar"}, annotations={"summary": "bar is meh"})
        ]
        self.assertEqual(1, len(am._get_alerts()))

        # Pretend there are no firing alerts now, thus there must be only one resolved alert being sent
        current_alerts_mock.return_value = []
        alerts = am._get_alerts()
        self.assertEqual(1, len(alerts))
        self.assertIsInstance(alerts[0].ends_at, datetime.datetime)
        self.assertIsInstance(alerts[0].ends_at.tzinfo, datetime.timezone)

        # Next iteration still no alerts, no alerts to send
        self.assertEqual(0, len(am._get_alerts()))


class TestTeamTagMatcher(unittest.TestCase):
    def test_raises_if_bad_yaml(self):
        with self.assertRaises(Exception):
            TeamTagMatcher.from_team_regexes_file(
                team_regexes_file=io.StringIO("I'm not really yaml")
            )

    def test_compiles_the_regexes_if_correct_yaml_with_alertname(self):
        my_team_tag_matcher = TeamTagMatcher.from_team_regexes_file(
            team_regexes_file=io.StringIO(
                yaml.dump({"team1": {"alertname": [".*team1.*"]}})
            )
        )

        self.assertIn("team1", my_team_tag_matcher.teams_regexes)
        for compiled_regex in chain(
            my_team_tag_matcher.teams_regexes["team1"].instance,
            my_team_tag_matcher.teams_regexes["team1"].alertname,
        ):
            self.assertIsInstance(compiled_regex, re.Pattern)

    def test_compiles_the_regexes_if_correct_yaml_with_instance(self):
        my_team_tag_matcher = TeamTagMatcher.from_team_regexes_file(
            team_regexes_file=io.StringIO(
                yaml.dump({"team1": {"instance": [".*team1.*"]}})
            )
        )

        self.assertIn("team1", my_team_tag_matcher.teams_regexes)
        for compiled_regex in chain(
            my_team_tag_matcher.teams_regexes["team1"].instance,
            my_team_tag_matcher.teams_regexes["team1"].alertname,
        ):
            self.assertIsInstance(compiled_regex, re.Pattern)

    def test_raises_if_unknown_label(self):
        team_regexes_file = io.StringIO(
            yaml.dump({"team1": {"unknown label": [".*team1.*"]}})
        )

        with self.assertRaises(UnknownLabel):
            TeamTagMatcher.from_team_regexes_file(team_regexes_file=team_regexes_file)


class TestPrometheusIcingaAM(unittest.TestCase):
    @staticmethod
    def get_am(team_tag_matcher: TeamTagMatcher) -> PrometheusIcingaAM:
        notes_map = {
            re.compile("grafana"): "dashboard",
            re.compile("wiki"): "runbook",
            re.compile("logstash"): "logs",
        }
        my_am = PrometheusIcingaAM(
            alertmanagers=["http://localhost:42"],
            query_string="query_string",
            status_cgi_path="/path/status.cgi",
            notes_map=notes_map,
            team_tag_matcher=team_tag_matcher,
        )
        return my_am

    @staticmethod
    def get_stub_service_status(overrides: Dict[str, Any] = None):
        if overrides is None:
            service_status = {}
        else:
            service_status = overrides

        required_fields = [
            "host_name",
            "service_description",
            "status_information",
            "status",
        ]
        for required_field in required_fields:
            if required_field not in service_status:
                service_status[required_field] = f"dummy value for {required_field}"

        return service_status

    @classmethod
    def get_stub_status(
        cls, service_statuses_overrides: List[Dict[str, Any]]
    ) -> Dict[str, Any]:
        status = {
            "status": {
                "service_status": [
                    cls.get_stub_service_status(overrides=service_status_override)
                    for service_status_override in service_statuses_overrides
                ]
            }
        }
        return status

    def test_alerts_from_status_json_adds_team_tag_if_matching_alertname(self):
        my_team_tag_matcher = TeamTagMatcher(
            teams_regexes={
                "team1": TeamRegexes(alertname=[re.compile(".*")], instance=[]),
            }
        )
        my_am = self.get_am(
            team_tag_matcher=my_team_tag_matcher,
        )
        my_status = self.get_stub_status(
            service_statuses_overrides=[
                {"service_description": "Alert number 1"},
                {"service_description": "Alert number 2"},
            ]
        )

        gotten_alerts = my_am.alerts_from_status_json(status=my_status)

        for alert in gotten_alerts:
            self.assertIn("team", alert.labels)
            self.assertEqual(alert.labels["team"], "team1")

    def test_alerts_from_status_json_adds_team_tag_if_matching_instance(self):
        my_team_tag_matcher = TeamTagMatcher(
            teams_regexes={
                "team1": TeamRegexes(
                    alertname=[], instance=[re.compile(".*dummy_instance.*")]
                ),
            }
        )
        my_am = self.get_am(
            team_tag_matcher=my_team_tag_matcher,
        )
        my_status = self.get_stub_status(
            service_statuses_overrides=[
                {
                    "service_description": "Alert number 1",
                    "host_name": "dummy_instance1",
                },
                {
                    "service_description": "Alert number 2",
                    "host_name": "dummy_instance2",
                },
            ]
        )

        gotten_alerts = my_am.alerts_from_status_json(status=my_status)

        for alert in gotten_alerts:
            self.assertIn("team", alert.labels)
            self.assertEqual(alert.labels["team"], "team1")

    def test_alerts_from_status_json_adds_team_tag_only_if_matching(self):
        my_team_tag_matcher = TeamTagMatcher(
            teams_regexes={
                "team1": TeamRegexes(
                    alertname=[re.compile(".*should match.*")], instance=[]
                ),
            }
        )
        my_am = self.get_am(
            team_tag_matcher=my_team_tag_matcher,
        )
        my_status = self.get_stub_status(
            service_statuses_overrides=[
                {"service_description": "Alert that should match"},
                {"service_description": "Alert that should not match"},
            ]
        )

        gotten_alerts = my_am.alerts_from_status_json(status=my_status)

        self.assertIn("team", gotten_alerts[0].labels)
        self.assertEqual(gotten_alerts[0].labels["team"], "team1")
        self.assertNotIn("team", gotten_alerts[1].labels)

    def test_alerts_from_status_json_adds_team_tag_only_if_team_tag_matcher_passed(
        self,
    ):
        my_am = self.get_am(team_tag_matcher=None)
        my_status = self.get_stub_status(
            service_statuses_overrides=[{"service_description": "Alert 1"}]
        )

        gotten_alerts = my_am.alerts_from_status_json(status=my_status)

        self.assertNotIn("team", gotten_alerts[0].labels)

    def test_alerts_from_status_json_adds_only_first_matching_team_tag(self):
        my_team_tag_matcher = TeamTagMatcher(
            teams_regexes={
                "team1": TeamRegexes(
                    alertname=[re.compile(".*should match.*")], instance=[]
                ),
                "team2": TeamRegexes(
                    alertname=[re.compile(".*should match.*")], instance=[]
                ),
            }
        )
        my_am = self.get_am(
            team_tag_matcher=my_team_tag_matcher,
        )
        my_status = self.get_stub_status(
            service_statuses_overrides=[
                {"service_description": "Alert that should match"},
            ]
        )

        gotten_alerts = my_am.alerts_from_status_json(status=my_status)

        self.assertIn("team", gotten_alerts[0].labels)
        self.assertEqual(gotten_alerts[0].labels["team"], "team1")


class TestMain(unittest.TestCase):
    @patch(PrometheusIcingaExporterModuleName + ".threading.enumerate")
    @patch(PrometheusIcingaExporterModuleName + ".PrometheusIcingaAM")
    @patch(PrometheusIcingaExporterModuleName + ".start_http_server")
    def test_works_without_params(
        self, mock_start_http_server, mock_PrometheusIcingaAM, mock_threading_enumerate
    ):
        gotten_return_code = main(args=[])

        self.assertEqual(0, gotten_return_code)

    @patch(PrometheusIcingaExporterModuleName + ".threading.enumerate")
    @patch(PrometheusIcingaExporterModuleName + ".PrometheusIcingaAM")
    @patch(PrometheusIcingaExporterModuleName + ".start_http_server")
    def test_fails_when_passed_bad_params(
        self, mock_start_http_server, mock_PrometheusIcingaAM, mock_threading_enumerate
    ):
        with self.assertRaises(SystemExit):
            main(args=["--im_not_a_param"])

    @patch(PrometheusIcingaExporterModuleName + ".threading.enumerate")
    @patch(PrometheusIcingaExporterModuleName + ".PrometheusIcingaAM")
    @patch(PrometheusIcingaExporterModuleName + ".start_http_server")
    def test_listen_accepted_and_passed_to_start_http_server(
        self, mock_start_http_server, mock_PrometheusIcingaAM, mock_threading_enumerate
    ):
        expected_port = 1234
        expected_addr = "0.0.0.0"

        gotten_return_code = main(args=["--listen", "0.0.0.0:1234"])

        self.assertEqual(0, gotten_return_code)
        mock_start_http_server.assert_called_with(expected_port, addr=expected_addr)

    @patch(PrometheusIcingaExporterModuleName + ".logging.basicConfig")
    @patch(PrometheusIcingaExporterModuleName + ".threading.enumerate")
    @patch(PrometheusIcingaExporterModuleName + ".PrometheusIcingaAM")
    @patch(PrometheusIcingaExporterModuleName + ".start_http_server")
    def test_debug_accepted_and_passed_to_logging(
        self,
        mock_start_http_server,
        mock_PrometheusIcingaAM,
        mock_threading_enumerate,
        mock_logging_basicConfig,
    ):
        expected_port = 1234
        expected_addr = "0.0.0.0"

        gotten_return_code = main(args=["--debug"])

        self.assertEqual(0, gotten_return_code)
        mock_logging_basicConfig.assert_called_with(level=logging.DEBUG)

    @patch(PrometheusIcingaExporterModuleName + ".threading.enumerate")
    @patch(PrometheusIcingaExporterModuleName + ".PrometheusIcingaAM")
    @patch(PrometheusIcingaExporterModuleName + ".start_http_server")
    def test_status_querystring_accepted_and_passed_to_PrometheusIcingaAM(
        self, mock_start_http_server, mock_PrometheusIcingaAM, mock_threading_enumerate
    ):
        expected_query_string = "some_random&status=querystring"

        gotten_return_code = main(args=["--status.querystring", expected_query_string])

        self.assertEqual(0, gotten_return_code)
        mock_PrometheusIcingaAM.assert_called_with(
            ANY,
            expected_query_string,
            ANY,
            ANY,
            notes_map=ANY,
            team_tag_matcher=ANY,
            extinfo_url=ANY,
        )

    @patch(PrometheusIcingaExporterModuleName + ".threading.enumerate")
    @patch(PrometheusIcingaExporterModuleName + ".PrometheusIcingaAM")
    @patch(PrometheusIcingaExporterModuleName + ".start_http_server")
    def test_status_path_accepted_and_passed_to_PrometheusIcingaAM(
        self, mock_start_http_server, mock_PrometheusIcingaAM, mock_threading_enumerate
    ):
        expected_status_cgi_path = "/pat/to/status.cgi"

        gotten_return_code = main(args=["--status.path", expected_status_cgi_path])

        self.assertEqual(0, gotten_return_code)
        mock_PrometheusIcingaAM.assert_called_with(
            ANY,
            ANY,
            expected_status_cgi_path,
            ANY,
            notes_map=ANY,
            team_tag_matcher=ANY,
            extinfo_url=ANY,
        )

    @patch(PrometheusIcingaExporterModuleName + ".threading.enumerate")
    @patch(PrometheusIcingaExporterModuleName + ".PrometheusIcingaAM")
    @patch(PrometheusIcingaExporterModuleName + ".start_http_server")
    def test_interval_accepted_and_passed_to_PrometheusIcingaAM(
        self, mock_start_http_server, mock_PrometheusIcingaAM, mock_threading_enumerate
    ):
        expected_interval = 42

        gotten_return_code = main(args=["--interval", str(expected_interval)])

        self.assertEqual(0, gotten_return_code)
        mock_PrometheusIcingaAM.assert_called_with(
            ANY,
            ANY,
            ANY,
            expected_interval,
            notes_map=ANY,
            team_tag_matcher=ANY,
            extinfo_url=ANY,
        )

    @patch(PrometheusIcingaExporterModuleName + ".threading.enumerate")
    @patch(PrometheusIcingaExporterModuleName + ".PrometheusIcingaAM")
    @patch(PrometheusIcingaExporterModuleName + ".start_http_server")
    def test_alertmanager_url_accepted_and_passed_to_PrometheusIcingaAM(
        self, mock_start_http_server, mock_PrometheusIcingaAM, mock_threading_enumerate
    ):
        expected_alertmanagers = ["http://alrtmanager1", "http://alertmanager2"]

        gotten_return_code = main(
            args=[
                "--alertmanager.url",
                expected_alertmanagers[0],
                "--alertmanager.url",
                expected_alertmanagers[1],
            ]
        )

        self.assertEqual(0, gotten_return_code)
        mock_PrometheusIcingaAM.assert_called_with(
            expected_alertmanagers,
            ANY,
            ANY,
            ANY,
            notes_map=ANY,
            team_tag_matcher=ANY,
            extinfo_url=ANY,
        )

    @patch(PrometheusIcingaExporterModuleName + ".open")
    @patch(PrometheusIcingaExporterModuleName + ".threading.enumerate")
    @patch(PrometheusIcingaExporterModuleName + ".PrometheusIcingaAM")
    @patch(PrometheusIcingaExporterModuleName + ".start_http_server")
    def test_label_team_config_file_accepted_and_passed_to_open(
        self,
        mock_start_http_server,
        mock_PrometheusIcingaAM,
        mock_threading_enumerate,
        mock_open,
    ):
        expected_config_file = "/path/to/some/config.yml"
        mock_open.return_value = io.StringIO("---\n{}")

        gotten_return_code = main(
            args=["--labels.team.config-file", expected_config_file]
        )

        self.assertEqual(0, gotten_return_code)
        mock_open.assert_called_with(expected_config_file, "r")

    @patch(PrometheusIcingaExporterModuleName + ".threading.enumerate")
    @patch(PrometheusIcingaExporterModuleName + ".PrometheusIcingaAM")
    @patch(PrometheusIcingaExporterModuleName + ".start_http_server")
    def test_extinfo_url_accepted_and_passed_to_PrometheusIcingaAM(
        self, mock_start_http_server, mock_PrometheusIcingaAM, mock_threading_enumerate
    ):
        expected_extinfo_url = "https://icinga.foo/cgi/extinfo.cgi"

        gotten_return_code = main(args=["--icinga.extinfo-url", expected_extinfo_url])

        self.assertEqual(0, gotten_return_code)
        mock_PrometheusIcingaAM.assert_called_with(
            ANY,
            ANY,
            ANY,
            ANY,
            notes_map=ANY,
            team_tag_matcher=ANY,
            extinfo_url=expected_extinfo_url,
        )
